
#ifndef __memoryManager__
#define __memoryManager__

#include <iostream>
#include <fstream>
#include <stdio.h>

#include <myheader.h>

#include <yarp/os/RateThread.h>
#include <yarp/os/Semaphore.h>

/*
 * memoryError exception.
 */
class memoryError {
private:
    int code;
    
public:
    memoryError(int x) { code = x; }
    int print() { fprintf(stdout,"mem error code: %d\n",code); return code; }
};


/*
 * a simple memory manager class w/ pagination.
 */
class memoryManager : public yarp::os::RateThread {
protected:
    // page_table bit code: | x * * * x P M R |
    unsigned char *page_table;          // page table
    int mem_map;                        // memory map (a bitmap of size 8 would be more than enough)
    unsigned char *ram;                 // memory
    int lastReferencedPageIdx;          // self-explaining
    std::fstream file;                  // swap space
    yarp::os::Semaphore *mutex;         // mutex to safely access to page table
    
    /**
     * va2pa()
     * return the PA corresponding to a given VA.
     * a value of ADDRESS_SPACE means page fault.
     */
    unsigned char * va2pa(unsigned char *va, bool modified = false);

    /**
     * demandPage()
     * called only on page fault.
     * - if one entry in page_table is available then load memory from disk image
     * - if no entries are available then call replacement()
     * - load memory from disk image into position indicated by replacement()
     */
    bool demandPage(unsigned char *va);
         
    /**
     * releasePage(), TO BE IMPLEMENTED
     */     
    bool releasePage(unsigned char *va);
    
    /**
     * loadPageFromDisk()
     * - load a specific page from disk file (which is ordered, 64K) 
     * - index is the page index 
     * - va is a virtual address (which addresses the image of the process on disk)
     * - pa is a physical address of the start of the page
     * - 4Kb are transferred
     */
    bool loadPageFromDisk(const unsigned char index, unsigned char *va, unsigned char *pa);

    /**
     * savePageToDisk()
     * - save page to disk file (ordered, 64K) 
     * - index is the page index 
     * - va is the VA corresponding to the page that has been evicted
     * - pa is the PA of the start of the page in ram
     * - 4Kb are transferred
     */
    bool savePageToDisk(const unsigned char index, unsigned char *pa, unsigned char *va);
        
    /**
     * replacement()
     * - scans the page_table and finds a page to replace, returns the page_table entry
     */
    int replacement();

    /**
     * run()
     * - periodically resets the R bit for all pages in currently 
     *   available
     */
    virtual void run();

    /**
     * rwAddressHelper()
     * - helper function for memory r/w access
     */
    unsigned char * rwAddressHelper(unsigned char *va, const bool modified);

    /**
     * deref()
     * - return a valid Linux/Windows address in 'ram'
     */
    inline unsigned char * deref(unsigned char *address) {
        const int ad = (int)address;
        if (ad < AVAILABLE_PHYS_MEMORY) {
            return ram+ad;
        }
        else {
            return 0;
        }
    } 
    
public:
    memoryManager(const unsigned int period);
    virtual ~memoryManager();
    
    /**
     * readAddress()
     * - call va2pa
     * - if !pagefault, then return pa
     * - on pagefault, call demandPage()
     * - when new page in ram then return pa
     */
    const unsigned char * readAddress(unsigned char *va);

    /**
     * writeAddress()
     * - identical to readAddress() but sets also M bit to 1
     */
    unsigned char * writeAddress(unsigned char *va);
};


#endif
