
#include <memoryManager.h>
#include <myheader.h>
#include <memory.h>
#include <string>

using namespace yarp;
using namespace yarp::os;
using namespace std;

//
// constructor
//
memoryManager::memoryManager(const unsigned int period) : 
               RateThread(period), ram(0), page_table(0), mutex(0) {
    ram = new unsigned char[AVAILABLE_PHYS_MEMORY];
    page_table = new unsigned char[PAGETABLESIZE];
    mutex = new Semaphore(1);    
    mem_map = 0;
    lastReferencedPageIdx = -1;

    if (ram == 0 || page_table == 0 || mutex == 0) {
        throw memoryError(1000);
    }

    memset(page_table, 0, PAGETABLESIZE);

    // prepare a swap file whose content is preset to 0
    char buf[AVAILABLE_PHYS_MEMORY*MAX_PROCESSES];
    memset(buf, 0, sizeof(buf));

    ofstream fout(SWAP_FILENAME, ios::binary);
    fout.write(buf, sizeof(buf));
    fout.close();

    file.open(SWAP_FILENAME, ios::in | ios::out | ios::binary);
}

//
// destructor
//
memoryManager::~memoryManager() {
    if (ram) {
        delete[] ram;
        ram = 0;
    }

    if (page_table) {
        delete[] page_table;
        page_table = 0;
    }

    if (mutex) {
        delete mutex;
        mutex = 0;
    }

    if (file.is_open()) {
        file.close();
    }

    mem_map = 0;
}


// va2pa()
// page_table bit code: | x * * * x P M R |
//
unsigned char * memoryManager::va2pa(unsigned char *va, bool modified) {
    const unsigned char index = (int(va) & 0x0000f000) >> 12;
    const unsigned int offset = (int(va) & 0x00000fff);
    unsigned char& entry = page_table[index];
    const bool present = (entry & 0x04)>0;
    if (present) {
        if (modified) {
            entry |= 0x03;  // modified & referenced
        }
        else {
            entry |= 0x01;  // only referenced
        }
        return deref((unsigned char *)(((entry & 0x70) << 8) + offset));
    }
    return (unsigned char *)PAGE_FAULT;
}


//
// loadPageFromDisk()
//
bool memoryManager::loadPageFromDisk(const unsigned char index, unsigned char *va, unsigned char *pa) {
    fprintf(stdout,"MMU: page %d loaded from disk\n",index);
    char *dest = (char *)deref(pa);
    const int file_ptr = (int(va) & 0x0000f000);
    file.seekg(file_ptr, ios::beg);
    file.read(dest, PAGESIZE);
    return true;
}

//
// savePageToDisk()
//
bool memoryManager::savePageToDisk(const unsigned char index, unsigned char *pa, unsigned char *va) {
    fprintf(stdout,"MMU: page %d saved to disk\n",index);
    const char *src = (const char *)deref(pa);
    const int file_ptr = (int(va) & 0x0000f000);
    file.seekp(file_ptr, ios::beg);
    file.write(src, PAGESIZE);
    return true;
}

//
// demandPage()
//
bool memoryManager::demandPage(unsigned char *va) {
    int i;
    int j = 0x01;
    unsigned char *freed = 0;   // page in ram.
    unsigned char index = 0;    // index in page_table.

    fprintf(stdout,"MMU: page fault occured\n");
    
    // scan for a free page in ram.
    for (i = 0; i < MEMORYSIZE; i++) {
        if (!(mem_map & j)) {
            break;
        }
        j = j << 1;
    }
         
    // i is now the empty page or MEMORYSIZE (eviction required).
    if (i == MEMORYSIZE) {
        // run the page replacement algorithm (return index in page_table).
        index = replacement();
        fprintf(stdout,"MMU: replacing page %d ...\n",index);
        i = page_table[index] >> 4;
        
        // check if M == true
        if (page_table[index] & 0x02) {
            savePageToDisk(index, (unsigned char *)(i << 12), (unsigned char *)(index << 12));
            // reset also mem_map.
            j = ~(0x01 << i);
            mem_map &= j;   // set the memory map (new free page).
        }
        else {
            fprintf(stdout,"MMU: page %d unchanged => just discarded from RAM\n",index);
        }

        // modify page_table to take eviction into account.
        page_table[index] = 0;
    }

    // PA corresponding to the free page in ram.
    freed = (unsigned char *)(i << 12);
    j = 0x01 << i;
    mem_map |= j;   // set the memory map (new used page).
    
    // load page anyway to resolve the page fault.
    index = (int(va) & 0x0000f000) >> 12;
    loadPageFromDisk(index, va, freed);
    
    // update the page_table.    
    page_table[index] = (i << 4);
    page_table[index] |= 0x04; // make it present in ram.
    
    return true;
}

//
// replacement()
//
int memoryManager::replacement()
{
    unsigned char index;

    // the page_table should be randomly scanned, otherwise,
    // if we start from index=0, lower indexes are weighted
    // more than higher ones; this behaviour is also affected
    // by the R-clearing rate.

    // check page class 0
    for (index=0; index<PAGETABLESIZE; index++) {
        unsigned char entry = page_table[index];
        if ((entry&0x04) && !(entry&0x03)) {
            return index;
        }
    }

    // check page class 1
    for (index=0; index<PAGETABLESIZE; index++) {
        unsigned char entry = page_table[index];
        if ((entry&0x04) && !(entry&0x01) && (entry&0x02)) {
            return index;
        }
    }

    // check page class 2
    for (index=0; index<PAGETABLESIZE; index++) {
        unsigned char entry = page_table[index];
        if ((entry&0x04) && (entry&0x01) && !(entry&0x02)) {
            return index;
        }
    }

    // check page class 3
    for (index=0; index<PAGETABLESIZE; index++) {
        unsigned char entry = page_table[index];
        if ((entry&0x04) && (entry&0x01) && (entry&0x02)) {
            return index;
        }
    }

    return 0;   // this point shouldn't be reached.
}

//
// run()
//
void memoryManager::run() {
    mutex->wait();

    for (unsigned char index=0; index<PAGETABLESIZE; index++) {
        page_table[index]&=~0x01;   // clear the R bit.
    }

    mutex->post();
}

//
// rwAddressHelper()
//
unsigned char * memoryManager::rwAddressHelper(unsigned char *va, const bool modified) {
    mutex->wait();

    unsigned char *pa = va2pa(va, modified);
    if (pa == (unsigned char *)PAGE_FAULT) {
        demandPage(va);
        pa = va2pa(va); // must be in ram at this point (assuming demand + va2pa are atomic)
    }

    const unsigned char index = (int(va) & 0x0000f000) >> 12;
    if (index != lastReferencedPageIdx) {
        fprintf(stdout,"MMU: page %d referenced\n",index);
        lastReferencedPageIdx = index;
    }

    mutex->post();

    return pa;
}

//
// readAddress()
//
const unsigned char * memoryManager::readAddress(unsigned char *va) {
    return rwAddressHelper(va, false);
}

//
// writeAddress()
//
unsigned char * memoryManager::writeAddress(unsigned char *va) {
    return rwAddressHelper(va, true);
}


